@@ -5,40 +5,91 @@ import urllib3
 
 from .version import __version__
 
+from http.client import responses as http_responses
 from requests.auth import HTTPBasicAuth
 from requests.exceptions import ConnectionError
 from time import sleep
-from typing import Dict
-from uuid import UUID
+from typing import Dict, List
+from uuid import uuid4, UUID
 
+#: protocol used to connect to api
+API_PROTOCOL = 'https'
 
+#: user agent sent to fmc
+API_USER_AGENT = f'FireREST/{__version__}'
+
+#: url used to generate token for api authorization
 API_AUTH_URL = '/api/fmc_platform/v1/auth/generatetoken'
+
+#: url used to refresh existing authorization token
 API_REFRESH_URL = '/api/fmc_platform/v1/auth/refreshtoken'
+
+#: url used to access platform related api calls
 API_PLATFORM_URL = '/api/fmc_platform/v1'
+
+#: url used to access configuration related api calls
 API_CONFIG_URL = '/api/fmc_config/v1'
 
+#: content type. as of 6.5.0 FMC only supports json
+API_CONTENT_TYPE = 'application/json'
+
+#: paging limit for get requests that contain multiple items
+API_PAGING_LIMIT = 1000
+
+#: expansion mode for get requests
+API_EXPANSION_MODE = True
+
+#: http request timeout
+API_REQUEST_TIMEOUT = 120
+
+#: name of fmc default domain for api requests
+API_DEFAULT_DOMAIN = 'Global'
+
+#: retry timer in seconds for repeating failed api calls
+API_RETRY_TIMER = 10
+
 
 class FireRESTApiException(Exception):
+    '''
+    generic api error exception
+    '''
+
     def __init__(self, message):
         super().__init__(message)
 
 
 class FireRESTAuthException(Exception):
+    '''
+    generic api authentication failure exception
+    '''
+
     def __init__(self, message):
         super().__init__(message)
 
 
 class FireRESTAuthRefreshException(Exception):
+    '''
+    exception used when api token refresh fails
+    '''
+
     def __init__(self, message):
         super().__init__(message)
 
 
 class FireRESTRateLimitException(Exception):
+    '''
+    exception used when fmc rate limiter kicks in
+    '''
+
     def __init__(self, message):
         super().__init__(message)
 
 
 class RequestDebugDecorator(object):
+    '''
+    decorator that adds additional debug logging for api requests
+    '''
+
     def __init__(self, action):
         self.action = action
 
@@ -47,38 +98,46 @@ class RequestDebugDecorator(object):
             action = self.action
             logger = args[0].logger
             request = args[1]
-            logger.debug(f'{action}: {request}')
+            request_id = str(uuid4())[:8]
+            try:
+                data = args[3]
+            except IndexError:
+                data = None
+            logger.debug(f'[{request_id}] [{action}] {request}')
+            if data:
+                logger.debug(f'[{request_id}] [Data] {data}')
             result = f(*args)
             status_code = result.status_code
-            logger.debug(f'Response Code: {status_code}')
-            if status_code >= 400:
-                logger.debug(f'Error: {result.content}')
+            status_code_name = http_responses[status_code]
+            logger.debug(f'[{request_id}] [Response] {status_code_name} ({status_code})')
+            if status_code >= 299:
+                logger.debug(f'[{request_id}] [Message] {result.content}')
             return result
 
         return wrapped_f
 
 
 class Client(object):
-    def __init__(self, hostname: str, username: str, password: str, session=dict(), protocol='https',
-                 verify_cert=False, logger=None, domain='Global', timeout=120):
+    def __init__(self, hostname: str, username: str, password: str, session=None, protocol=API_PROTOCOL,
+                 verify_cert=False, logger=None, domain=API_DEFAULT_DOMAIN, timeout=API_REQUEST_TIMEOUT):
         '''
-        Initialize api client object
+        Initialize api client object (make sure to use a dedicated api user!)
         :param hostname: ip address or dns name of fmc
         :param username: fmc username
         :param password: fmc password
         :param session: authentication session (can be provided in case api client should not generate one at init).
                       Make sure to pass the headers of a successful authentication to the session variable,
                       otherwise this will fail
-        :param protocol: protocol used to access fmc api. default = https
-        :param verify_cert: check fmc certificate for vailidity. default = False
+        :param protocol: protocol used to access fmc api
+        :param verify_cert: check fmc certificate for vailidity
         :param logger: optional logger instance, in case debug logging is needed
-        :param domain: name of the fmc domain. default = Global
-        :param timeout: timeout value for http requests. default = 120
+        :param domain: name of the fmc domain
+        :param timeout: timeout value for http requests
         '''
         self.headers = {
-            'Content-Type': 'application/json',
-            'Accept': 'application/json',
-            'User-Agent': f'FireREST/{__version__}'
+            'Content-Type': API_CONTENT_TYPE,
+            'Accept': API_CONTENT_TYPE,
+            'User-Agent': API_USER_AGENT,
         }
         self.refresh_counter = 0
         self.logger = self._get_logger(logger)
@@ -106,16 +165,45 @@ class Client(object):
         :return: dummy logger instance if logger is None, otherwise return logger variable again
         '''
         if not logger:
-            dummy_logger = logging.getLogger('FireREST.Client')
+            dummy_logger = logging.getLogger(f'{API_USER_AGENT}.Client')
             dummy_logger.addHandler(logging.NullHandler())
             return dummy_logger
         return logger
 
-    def _is_getbyid_operation(self, request):
+    def _squash_responses(self, responses: List):
+        if len(responses) == 1:
+            payload = responses[0].json()
+
+            # get request did not yield any results
+            if not payload['links']:
+                return list()
+
+            # get request contains multiple results without paging
+            if 'items' in payload:
+                return payload['items']
+
+            # get request returned single item
+            return responses[0].json()
+        items = list()
+        for response in responses:
+            if 'items' in response:
+                items.extend(response.json()['items'])
+            else:
+                self.logger.debug('Response {response.url} did not contain any items. Skipping...')
+        return items
+
+    def _is_getbyid_operation(self, request: str):
+        '''
+        Verify if a get request contains multiple items
+        :param request: request url
+        :return: True if operation returns a single item, False if it contains multiple items
+        '''
         try:
             val = UUID(request.split('/')[-1])  # noqa: F841
             return True
         except ValueError:
+            if 'overrides' in request:
+                return True
             return False
 
     def _url(self, namespace='base', path=str()):
@@ -125,21 +213,21 @@ class Client(object):
         :param path: the url path for which a full url should be created
         :return: url in string format
         '''
-        if namespace == 'config':
-            return f'{self.protocol}://{self.hostname}{API_CONFIG_URL}/domain/{self.domain}{path}'
-        if namespace == 'platform':
-            return f'{self.protocol}://{self.hostname}{API_PLATFORM_URL}{path}'
-        if namespace == 'auth':
-            return f'{self.protocol}://{self.hostname}{API_AUTH_URL}'
-        if namespace == 'refresh':
-            return f'{self.protocol}://{self.hostname}{API_REFRESH_URL}'
-        return f'{self.protocol}://{self.hostname}{path}'
+        options = {
+            'base': f'{self.protocol}://{self.hostname}{path}',
+            'config': f'{self.protocol}://{self.hostname}{API_CONFIG_URL}/domain/{self.domain}{path}',
+            'platform': f'{self.protocol}://{self.hostname}{API_PLATFORM_URL}{path}',
+            'refresh': f'{self.protocol}://{self.hostname}{API_REFRESH_URL}',
+        }
+        if namespace not in options.keys():
+            raise ValueError(f'Invalid namespace "{namespace}" provided. Valid namespaces: {options.keys()}')
+        return options[namespace]
 
     def _login(self):
         '''
         Login to fmc api and save X-auth-access-token, X-auth-refresh-token and DOMAINS to variables
         '''
-        request = self._url('auth')
+        request = f'{self.protocol}://{self.hostname}{API_AUTH_URL}'
         try:
             response = requests.post(request, headers=self.headers, auth=self.cred, verify=self.verify_cert)
 
@@ -148,9 +236,8 @@ class Client(object):
                 raise FireRESTAuthException(f'API Authentication to {self.hostname} failed.')
 
             if response.status_code == 429:
-                msg = f'API Authentication to {self.hostname} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'API Authentication to {self.hostname} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
-
             access_token = response.headers.get('X-auth-access-token', default=None)
             refresh_token = response.headers.get('X-auth-refresh-token', default=None)
             if not access_token or not refresh_token:
@@ -166,8 +253,9 @@ class Client(object):
             self.logger.error(exc, exc_info=True)
             raise
         except FireRESTRateLimitException:
-            self.logger.debug(f'Could not login to {self.hostname}. Rate limit exceeded. Backing of for 10 seconds.')
-            sleep(10)
+            self.logger.debug(
+                f'Could not login to {self.hostname}. Rate limit exceeded. Retrying in {API_RETRY_TIMER} seconds.')
+            sleep(API_RETRY_TIMER)
             self._login()
 
     def _refresh(self):
@@ -186,7 +274,7 @@ class Client(object):
             response = requests.post(request, headers=self.headers, verify=self.verify_cert)
 
             if response.status_code == 429:
-                msg = f'API token refresh to {self.hostname} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'API token refresh to {self.hostname} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
 
             access_token = response.headers.get('X-auth-access-token', default=None)
@@ -202,8 +290,8 @@ class Client(object):
             raise
         except FireRESTRateLimitException:
             self.logger.debug(
-                f'API token refresh to {self.hostname} failed. Rate limit exceeded. Backing of for 10 seconds.')
-            sleep(10)
+                f'API token refresh to {self.hostname} failed. Rate limit exceeded. Retrying in {API_RETRY_TIMER} seconds.')
+            sleep(API_RETRY_TIMER)
             self._login()
         except FireRESTApiException as exc:
             self.logger.error(str(exc))
@@ -229,63 +317,62 @@ class Client(object):
                     self._refresh()
                     return self._delete(request, params)
             if response.status_code == 429:
-                msg = f'DELETE operation {request} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'DELETE operation {request} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
         except FireRESTRateLimitException:
-            sleep(10)
+            sleep(API_RETRY_TIMER)
             return self._delete(request, params)
         return response
 
     @RequestDebugDecorator('GET')
-    def _get_request(self, request: str, params=None, limit=100):
+    def _get_request(self, request: str, params=None):
         '''
         GET Operation for FMC REST API. In case of authentication issues session will be refreshed
         :param request: URL of request that should be performed
         :param params: dict of parameters for http request
-        :param limit: set custom limit for paging. If not set, api will default to 100
         :return: requests.Response object
         '''
         if params is None:
             params = dict()
-        if not self._is_getbyid_operation(request):
-            params['limit'] = limit
+        if not self._is_getbyid_operation(request) and 'limit' not in params:
+            params['limit'] = API_PAGING_LIMIT
         try:
             response = requests.get(request, headers=self.headers, params=params, verify=self.verify_cert,
                                     timeout=self.timeout)
             if response.status_code == 401:
                 if 'Access token invalid' in str(response.json()):
                     self._refresh()
-                    return self._get_request(request, params, limit)
+                    return self._get_request(request, params)
             if response.status_code == 429:
-                msg = f'GET operation {request} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'GET operation {request} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
         except FireRESTRateLimitException:
-            sleep(10)
-            return self._get_request(request, params, limit)
+            sleep(API_RETRY_TIMER)
+            return self._get_request(request, params)
         return response
 
-    def _get(self, request: str, params=None, limit=100):
+    def _get(self, request: str, params=None):
         '''
         GET Operation that supports paging for FMC REST API. In case of authentication issues session will be refreshed
         :param request: URL of request that should be performed
         :param params: dict of parameters for http request
-        :param limit: set custom limit for paging. If not set, api will default to 100
         :return: list of requests.Response objects
         '''
         if params is None:
             params = dict()
         responses = list()
-        response = self._get_request(request, params, limit)
+        response = self._get_request(request, params)
         responses.append(response)
         payload = response.json()
         if 'paging' in payload.keys():
             pages = int(payload['paging']['pages'])
-            limit = int(payload['paging']['limit'])
+            params['limit'] = int(payload['paging']['limit'])
             for i in range(1, pages, 1):
-                params['offset'] = str(int(i) * limit)
-                response_page = self._get_request(request, params, limit)
+                params['offset'] = i * params['limit']
+                response_page = self._get_request(request, params)
                 responses.append(response_page)
-        return responses
+        print(self._squash_responses(responses))
+        return self._squash_responses(responses)
 
     @RequestDebugDecorator('POST')
     def _post(self, request: str, data: Dict, params=None):
@@ -306,10 +393,10 @@ class Client(object):
                     self._refresh()
                     return self._post(request, data, params)
             if response.status_code == 429:
-                msg = f'POST operation {request} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'POST operation {request} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
         except FireRESTRateLimitException:
-            sleep(10)
+            sleep(API_RETRY_TIMER)
             return self._post(request, data, params)
         return response
 
@@ -332,10 +419,10 @@ class Client(object):
                     self._refresh()
                     return self._put(request, data, params)
             if response.status_code == 429:
-                msg = f'PUT operation {request} failed due to FMC rate limiting. Backing off for 10 seconds.'
+                msg = f'PUT operation {request} failed due to FMC rate limiting. Retrying in {API_RETRY_TIMER} seconds.'
                 raise FireRESTRateLimitException(msg)
         except FireRESTRateLimitException:
-            sleep(10)
+            sleep(API_RETRY_TIMER)
             return self._put(request, data, params)
         return response
 
@@ -532,19 +619,35 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_objects(self, object_type: str, expanded=False):
+    def get_objects(self, object_type: str, expanded=API_EXPANSION_MODE):
         request = f'/object/{object_type}'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
+    def get_objects_override(self, objects: List, object_type: str, expanded=API_EXPANSION_MODE):
+        overrides = list()
+        for obj in objects:
+            if obj['overridable']:
+                responses = self.get_object_override(object_type, obj['id'], expanded=expanded)
+                overrides.extend(responses)
+        return overrides
+
     def get_object(self, object_type: str, object_id: str):
         request = f'/object/{object_type}/{object_id}'
         url = self._url('config', request)
         return self._get(url)
 
+    def get_object_override(self, object_type: str, object_id: str, expanded=API_EXPANSION_MODE):
+        request = f'/object/{object_type}/{object_id}/overrides'
+        url = self._url('config', request)
+        params = {
+            'expanded': expanded,
+        }
+        return self._get(url, params)
+
     def update_object(self, object_type: str, object_id: str, data: Dict):
         request = f'/object/{object_type}/{object_id}'
         url = self._url('config', request)
@@ -560,11 +663,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_devices(self, expanded=False):
+    def get_devices(self, expanded=API_EXPANSION_MODE):
         request = '/devices/devicerecords'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -583,10 +686,10 @@ class Client(object):
         url = self._url('config', request)
         return self._delete(url)
 
-    def get_device_hapairs(self, expanded=False):
+    def get_device_hapairs(self, expanded=API_EXPANSION_MODE):
         request = '/devicehapairs/ftddevicehapairs'
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         url = self._url('config', request)
         return self._get(url, params)
@@ -611,10 +714,10 @@ class Client(object):
         url = self._url('config', request)
         return self._delete(url)
 
-    def get_device_hapair_monitoredinterfaces(self, device_hapair_id: str, expanded=False):
+    def get_device_hapair_monitoredinterfaces(self, device_hapair_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devicehapairs/ftddevicehapairs/{device_hapair_id}/monitoredinterfaces'
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         url = self._url('config', request)
         return self._get(url, params)
@@ -629,11 +732,11 @@ class Client(object):
         url = self._url('config', request)
         return self._put(url, data)
 
-    def get_ftd_physical_interfaces(self, device_id: str, expanded=False):
+    def get_ftd_physical_interfaces(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/physicalinterfaces'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -652,11 +755,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_ftd_redundant_interfaces(self, device_id: str, expanded=False):
+    def get_ftd_redundant_interfaces(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/redundantinterfaces'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -680,11 +783,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_ftd_portchannel_interfaces(self, device_id: str, expanded=False):
+    def get_ftd_portchannel_interfaces(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/etherchannelinterfaces'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -708,11 +811,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_ftd_sub_interfaces(self, device_id: str, expanded=False):
+    def get_ftd_sub_interfaces(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/subinterfaces'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -736,11 +839,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_ftd_ipv4staticroutes(self, device_id: str, expanded=False):
+    def get_ftd_ipv4staticroutes(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/routing/ipv4staticroutes'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -764,11 +867,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_ftd_ipv6staticroutes(self, device_id: str, expanded=False):
+    def get_ftd_ipv6staticroutes(self, device_id: str, expanded=API_EXPANSION_MODE):
         request = f'/devices/devicerecords/{device_id}/routing/ipv6staticroutes'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -792,11 +895,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_deployable_devices(self, expanded=False):
+    def get_deployable_devices(self, expanded=API_EXPANSION_MODE):
         request = '/deployment/deployabledevices'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -805,18 +908,18 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_policies(self, policy_type: str, expanded=False):
+    def get_policies(self, policy_type: str, expanded=API_EXPANSION_MODE):
         request = f'/policy/{policy_type}'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
-    def get_policy(self, policy_id: str, policy_type: str, expanded=False):
+    def get_policy(self, policy_id: str, policy_type: str, expanded=API_EXPANSION_MODE):
         request = f'/policy/{policy_type}/{policy_id}'
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         url = self._url('config', request)
         return self._get(url, params)
@@ -860,16 +963,16 @@ class Client(object):
         url = self._url('config', request)
         return self._get(url)
 
-    def get_acp_rules(self, policy_id: str, expanded=False):
+    def get_acp_rules(self, policy_id: str, expanded=API_EXPANSION_MODE):
         request = f'/policy/accesspolicies/{policy_id}/accessrules'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
     def update_acp_rule(self, policy_id: str, rule_id: str, data: Dict):
-        request = '/policy/accesspolicies/{policy_id}/accessrules/{rule_id}'
+        request = f'/policy/accesspolicies/{policy_id}/accessrules/{rule_id}'
         url = self._url('config', request)
         return self._put(url, data)
 
@@ -888,11 +991,11 @@ class Client(object):
         url = self._url('config', request)
         return self._get(url)
 
-    def get_autonat_rules(self, policy_id: str, expanded=False):
+    def get_autonat_rules(self, policy_id: str, expanded=API_EXPANSION_MODE):
         request = f'/policy/ftdnatpolicies/{policy_id}/autonatrules'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -916,11 +1019,11 @@ class Client(object):
         url = self._url('config', request)
         return self._get(url)
 
-    def get_manualnat_rules(self, policy_id: str, expanded=False):
+    def get_manualnat_rules(self, policy_id: str, expanded=API_EXPANSION_MODE):
         request = f'/policy/ftdnatpolicies/manualnatrules/{policy_id}'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
@@ -939,11 +1042,11 @@ class Client(object):
         url = self._url('config', request)
         return self._post(url, data)
 
-    def get_policy_assignments(self, expanded=False):
+    def get_policy_assignments(self, expanded=API_EXPANSION_MODE):
         request = '/assignment/policyassignments'
         url = self._url('config', request)
         params = {
-            'expanded': expanded
+            'expanded': expanded,
         }
         return self._get(url, params)
 
